"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseErrorMessage = exports.waitForInitialization = void 0;
const logs = require("../logs");
/**
 * Periodically checks for the existence of a dataset and table until both are found or a maximum number of attempts is reached.
 * @param {WaitForInitializationParams} params - Parameters for initialization including the dataset and the table name to check.
 * @param {number} maxAttempts - Maximum number of attempts before giving up (defaults to 12).
 * @returns {Promise<Table>} A promise that resolves with the Table if it exists, or rejects if it doesn't exist after maxAttempts or an error occurs.
 * @throws {Error} Throws an error if the dataset or table cannot be verified to exist after multiple attempts or if an unexpected error occurs.
 */
async function waitForInitialization({ dataset, changelogName }, maxAttempts = 12) {
    return new Promise((resolve, reject) => {
        let attempts = 0;
        let handle = setInterval(async () => {
            try {
                const [datasetExists] = await dataset.exists();
                const table = dataset.table(changelogName);
                const [tableExists] = await table.exists();
                if (datasetExists && tableExists) {
                    clearInterval(handle);
                    resolve(table);
                }
                else {
                    attempts++;
                    if (attempts >= maxAttempts) {
                        clearInterval(handle);
                        reject(new Error("Initialization timed out. Dataset or table could not be verified to exist after multiple attempts."));
                    }
                }
            }
            catch (error) {
                clearInterval(handle);
                const message = error instanceof Error
                    ? error.message
                    : "An unexpected error occurred";
                logs.failedToInitializeWait(message);
                reject(new Error(message));
            }
        }, 500);
    });
}
exports.waitForInitialization = waitForInitialization;
function parseErrorMessage(error, process = "") {
    let message;
    if (error instanceof Error) {
        // Standard error handling
        message = error.message;
    }
    else if (typeof error === "object" &&
        error !== null &&
        "message" in error) {
        // Handling errors from APIs or other sources that are not native Error objects
        message = error.message;
    }
    else {
        // Fallback for when error is neither an Error object nor an expected structured object
        message = `An unexpected error occurred${process ? ` during ${process}` : ""}.`;
    }
    return message;
}
exports.parseErrorMessage = parseErrorMessage;
